package com.sun.java.util.jar.pack;

import java.io.FileInputStream;
import java.io.File;
import java.util.Collections;
import java.util.ArrayList;
import java.io.BufferedInputStream;
import java.io.InputStream;
import java.util.jar.JarEntry;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.util.ListIterator;
import java.util.Iterator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.beans.PropertyChangeListener;
import java.util.jar.JarInputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.jar.JarFile;
import java.util.SortedMap;
import java.util.jar.Pack200;

public class PackerImpl extends TLGlobals implements Pack200.Packer
{
    public PackerImpl() {
    }
    
    public SortedMap<String, String> properties() {
        return this.props;
    }
    
    public synchronized void pack(JarFile jarFile, OutputStream outputStream) throws IOException {
        assert Utils.currentInstance.get() == null;
        int n = !this.props.getBoolean("com.sun.java.util.jar.pack.default.timezone");
        try {
            Utils.currentInstance.set(this);
            if (!(n == 0)) {
                Utils.changeDefaultTimeZoneToUtc();
            }
            if ("0".equals(this.props.getProperty("pack.effort"))) {
                Utils.copyJarFile(jarFile, outputStream);
            }
            else {
                new PackerImpl.DoPack().run(jarFile, outputStream);
            }
        }
        finally {
            Utils.currentInstance.set(null);
            if (!(n == 0)) {
                Utils.restoreDefaultTimeZone();
            }
            jarFile.close();
        }
        return;
    }
    
    public synchronized void pack(JarInputStream jarInputStream, OutputStream outputStream) throws IOException {
        assert Utils.currentInstance.get() == null;
        int n = !this.props.getBoolean("com.sun.java.util.jar.pack.default.timezone");
        try {
            Utils.currentInstance.set(this);
            if (!(n == 0)) {
                Utils.changeDefaultTimeZoneToUtc();
            }
            if ("0".equals(this.props.getProperty("pack.effort"))) {
                Utils.copyJarFile(jarInputStream, outputStream);
            }
            else {
                new PackerImpl.DoPack().run(jarInputStream, outputStream);
            }
        }
        finally {
            Utils.currentInstance.set(null);
            if (!(n == 0)) {
                Utils.restoreDefaultTimeZone();
            }
            jarInputStream.close();
        }
        return;
    }
    
    public void addPropertyChangeListener(PropertyChangeListener propertyChangeListener) {
        this.props.addListener(propertyChangeListener);
    }
    
    public void removePropertyChangeListener(PropertyChangeListener propertyChangeListener) {
        this.props.removeListener(propertyChangeListener);
    }
    
    private class DoPack
    {
        final int verbose;
        final Package pkg;
        final String unknownAttrCommand;
        final String classFormatCommand;
        final Map<Attribute.Layout, Attribute> attrDefs;
        final Map<Attribute.Layout, String> attrCommands;
        final boolean keepFileOrder;
        final boolean keepClassOrder;
        final boolean keepModtime;
        final boolean latestModtime;
        final boolean keepDeflateHint;
        long totalOutputSize;
        int segmentCount;
        long segmentTotalSize;
        long segmentSize;
        final long segmentLimit;
        final List<String> passFiles;
        private int nread;
        static final /* synthetic */ boolean $assertionsDisabled;
        
        private DoPack() {
            this.verbose = PackerImpl.this.props.getInteger("com.sun.java.util.jar.pack.verbose");
            PackerImpl.this.props.setInteger("pack.progress", 0);
            if (this.verbose > 0) {
                Utils.log.info(PackerImpl.this.props.toString());
            }
            this.pkg = new Package(Package.Version.makeVersion(PackerImpl.this.props, "min.class"), Package.Version.makeVersion(PackerImpl.this.props, "max.class"), Package.Version.makeVersion(PackerImpl.this.props, "package"));
            String property = PackerImpl.this.props.getProperty("pack.unknown.attribute", "pass");
            if (!"strip".equals(property) && !"pass".equals(property) && !"error".equals(property)) {
                throw new RuntimeException("Bad option: pack.unknown.attribute = " + property);
            }
            this.unknownAttrCommand = property.intern();
            String property2 = PackerImpl.this.props.getProperty("com.sun.java.util.jar.pack.class.format.error", "pass");
            if (!"pass".equals(property2) && !"error".equals(property2)) {
                throw new RuntimeException("Bad option: com.sun.java.util.jar.pack.class.format.error = " + property2);
            }
            this.classFormatCommand = property2.intern();
            HashMap<Attribute.Layout, Attribute> hashMap = new HashMap();
            HashMap<Attribute.Layout, String> hashMap2 = new HashMap();
            String[] array = { "pack.class.attribute.", "pack.field.attribute.", "pack.method.attribute.", "pack.code.attribute." };
            int[] array2 = { 0, 1, 2, 3 };
            for (int i = 0; i < array2.length; ++i) {
                String prefix = array[i];
                for (String s : PackerImpl.this.props.prefixMap(prefix).keySet()) {
                    assert !(!s.startsWith(prefix));
                    String substring = s.substring(prefix.length());
                    String property3 = PackerImpl.this.props.getProperty(s);
                    Attribute.Layout keyForLookup = Attribute.keyForLookup(array2[i], substring);
                    if ("strip".equals(property3) || "pass".equals(property3) || "error".equals(property3)) {
                        hashMap2.put(keyForLookup, property3.intern());
                    }
                    else {
                        Attribute.define(hashMap, array2[i], substring, property3);
                        if (this.verbose > 1) {
                            Utils.log.fine("Added layout for " + Constants.ATTR_CONTEXT_NAME[i] + " attribute " + substring + " = " + property3);
                        }
                        assert !(!hashMap.containsKey(keyForLookup));
                        continue;
                    }
                }
            }
            this.attrDefs = (hashMap.isEmpty() ? null : hashMap);
            this.attrCommands = (hashMap2.isEmpty() ? null : hashMap2);
            this.keepFileOrder = PackerImpl.this.props.getBoolean("pack.keep.file.order");
            this.keepClassOrder = PackerImpl.this.props.getBoolean("com.sun.java.util.jar.pack.keep.class.order");
            this.keepModtime = "keep".equals(PackerImpl.this.props.getProperty("pack.modification.time"));
            this.latestModtime = "latest".equals(PackerImpl.this.props.getProperty("pack.modification.time"));
            this.keepDeflateHint = "keep".equals(PackerImpl.this.props.getProperty("pack.deflate.hint"));
            if (!this.keepModtime && !this.latestModtime) {
                int time = PackerImpl.this.props.getTime("pack.modification.time");
                if (time != 0) {
                    this.pkg.default_modtime = time;
                }
            }
            if (!this.keepDeflateHint && PackerImpl.this.props.getBoolean("pack.deflate.hint")) {
                Package pkg = this.pkg;
                pkg.default_options |= 0x20;
            }
            this.totalOutputSize = 0L;
            this.segmentCount = 0;
            this.segmentTotalSize = 0L;
            this.segmentSize = 0L;
            long long1;
            if (PackerImpl.this.props.getProperty("pack.segment.limit", "").equals("")) {
                long1 = -1L;
            }
            else {
                long1 = PackerImpl.this.props.getLong("pack.segment.limit");
            }
            long max = Math.max(-1L, Math.min(2147483647L, long1));
            if (max == -1L) {
                max = Long.MAX_VALUE;
            }
            this.segmentLimit = max;
            this.passFiles = PackerImpl.this.props.getProperties("pack.pass.file.");
            ListIterator<String> listIterator = this.passFiles.listIterator();
            while (listIterator.hasNext()) {
                String s2 = (String)listIterator.next();
                if (s2 == null) {
                    listIterator.remove();
                }
                else {
                    String s3 = Utils.getJarEntryName(s2);
                    if (s3.endsWith("/")) {
                        s3 = s3.substring(0, s3.length() - 1);
                    }
                    listIterator.set(s3);
                }
            }
            if (this.verbose > 0) {
                Utils.log.info("passFiles = " + this.passFiles);
            }
            int integer = PackerImpl.this.props.getInteger("com.sun.java.util.jar.pack.archive.options");
            if (integer != 0) {
                Package pkg2 = this.pkg;
                pkg2.default_options |= integer;
            }
            this.nread = 0;
            return;
        }
        
        boolean isClassFile(String substring) {
            if (!substring.endsWith(".class")) {
                return false;
            }
            while (!this.passFiles.contains(substring)) {
                int lastIndex = substring.lastIndexOf(47);
                if (lastIndex < 0) {
                    return true;
                }
                substring = substring.substring(0, lastIndex);
            }
            return false;
        }
        
        boolean isMetaInfFile(String s) {
            return s.startsWith("/META-INF") || s.startsWith("META-INF");
        }
        
        private void makeNextPackage() {
            this.pkg.reset();
        }
        
        private void noteRead(PackerImpl.DoPack.InFile inFile) {
            ++this.nread;
            if (this.verbose > 2) {
                Utils.log.fine("...read " + inFile.name);
            }
            if (this.verbose > 0 && this.nread % 1000 == 0) {
                Utils.log.info("Have read " + this.nread + " files...");
            }
        }
        
        void run(JarInputStream jarInputStream, OutputStream outputStream) throws IOException {
            if (jarInputStream.getManifest() != null) {
                ByteArrayOutputStream out = new ByteArrayOutputStream();
                jarInputStream.getManifest().write(out);
                this.pkg.addFile(this.readFile("META-INF/MANIFEST.MF", new ByteArrayInputStream(out.toByteArray())));
            }
            JarEntry nextJarEntry;
            while ((nextJarEntry = jarInputStream.getNextJarEntry()) != null) {
                PackerImpl.DoPack.InFile o = new PackerImpl.DoPack.InFile(nextJarEntry);
                String name = o.name;
                Package.File file = this.readFile(name, jarInputStream);
                file = null;
                long n = this.isMetaInfFile(name) ? 0L : o.getInputLength();
                long segmentSize = this.segmentSize + n;
                this.segmentSize = segmentSize;
                if (segmentSize > this.segmentLimit) {
                    this.segmentSize -= n;
                    this.flushPartial(outputStream, -1);
                }
                if (this.verbose > 1) {
                    Utils.log.fine("Reading " + name);
                }
                assert nextJarEntry.isDirectory() == name.endsWith("/");
                if (this.isClassFile(name)) {
                    file = this.readClass(name, file.getInputStream());
                }
                if (file == null) {
                    this.pkg.addFile(file);
                }
                o.copyTo(file);
                this.noteRead(o);
            }
            this.flushAll(outputStream);
        }
        
        void run(JarFile jarFile, OutputStream outputStream) throws IOException {
            List<PackerImpl.DoPack.InFile> scanJar = this.scanJar(jarFile);
            if (this.verbose > 0) {
                Utils.log.info("Reading " + scanJar.size() + " files...");
            }
            int n = 0;
            for (PackerImpl.DoPack.InFile inFile : scanJar) {
                String name = inFile.name;
                long n2 = this.isMetaInfFile(name) ? 0L : inFile.getInputLength();
                long segmentSize = this.segmentSize + n2;
                this.segmentSize = segmentSize;
                if (segmentSize > this.segmentLimit) {
                    this.segmentSize -= n2;
                    float n3 = (float)(n + 1);
                    float f = ((float)scanJar.size() - n3) * ((float)(this.segmentCount + 1) / n3);
                    if (this.verbose > 1) {
                        Utils.log.fine("Estimated segments to do: " + f);
                    }
                    this.flushPartial(outputStream, (int)Math.ceil((double)f));
                }
                InputStream inputStream = inFile.getInputStream();
                if (this.verbose > 1) {
                    Utils.log.fine("Reading " + name);
                }
                Package.File file = null;
                if (this.isClassFile(name)) {
                    file = this.readClass(name, inputStream);
                    if (file == null) {
                        inputStream.close();
                        inputStream = inFile.getInputStream();
                    }
                }
                if (file == null) {
                    file = this.readFile(name, inputStream);
                    this.pkg.addFile(file);
                }
                inFile.copyTo(file);
                inputStream.close();
                this.noteRead(inFile);
                ++n;
            }
            this.flushAll(outputStream);
        }
        
        Package.File readClass(String s, InputStream in) throws IOException {
            Package.Class class1 = this.pkg.new Package.Class(s);
            ClassReader classReader = new ClassReader(class1, new BufferedInputStream(in));
            classReader.setAttrDefs(this.attrDefs);
            classReader.setAttrCommands(this.attrCommands);
            classReader.unknownAttrCommand = this.unknownAttrCommand;
            try {
                classReader.read();
            }
            catch (IOException ex) {
                String s2 = "Passing class file uncompressed due to";
                if ((ex instanceof Attribute.FormatException)) {
                    Attribute.FormatException ex3 = (Attribute.FormatException)ex;
                    if (ex3.layout.equals("pass")) {
                        Utils.log.info(ex3.toString());
                        Utils.log.warning(s2 + " unrecognized attribute: " + s);
                        return null;
                    }
                }
                else if ((ex instanceof ClassReader.ClassFormatException)) {
                    ClassReader.ClassFormatException ex2 = (ClassReader.ClassFormatException)ex;
                    if (this.classFormatCommand.equals("pass")) {
                        Utils.log.info(ex2.toString());
                        Utils.log.warning(s2 + " unknown class format: " + s);
                        return null;
                    }
                }
                throw ex;
            }
            this.pkg.addClass(class1);
            return class1.file;
        }
        
        Package.File readFile(String s, InputStream inputStream) throws IOException {
            Package.File file = this.pkg.new Package.File(s);
            file.readFrom(inputStream);
            if (file.isDirectory() && file.getFileLength() != 0L) {
                throw new IllegalArgumentException("Non-empty directory: " + file.getFileName());
            }
            return file;
        }
        
        void flushPartial(OutputStream outputStream, int b) throws IOException {
            if (this.pkg.files.isEmpty() && this.pkg.classes.isEmpty()) {
                return;
            }
            this.flushPackage(outputStream, Math.max(1, b));
            PackerImpl.this.props.setInteger("pack.progress", 25);
            this.makeNextPackage();
            ++this.segmentCount;
            this.segmentTotalSize += this.segmentSize;
            this.segmentSize = 0L;
            return;
        }
        
        void flushAll(OutputStream outputStream) throws IOException {
            PackerImpl.this.props.setInteger("pack.progress", 50);
            this.flushPackage(outputStream, 0);
            outputStream.flush();
            PackerImpl.this.props.setInteger("pack.progress", 100);
            ++this.segmentCount;
            this.segmentTotalSize += this.segmentSize;
            this.segmentSize = 0L;
            if (this.verbose > 0 && this.segmentCount > 1) {
                Utils.log.info("Transmitted " + this.segmentTotalSize + " input bytes in " + this.segmentCount + " segments totaling " + this.totalOutputSize + " bytes");
            }
        }
        
        void flushPackage(OutputStream outputStream, int archiveNextCount) throws IOException {
            int size = this.pkg.files.size();
            if (!this.keepFileOrder) {
                if (this.verbose > 1) {
                    Utils.log.fine("Reordering files.");
                }
                this.pkg.reorderFiles(this.keepClassOrder, true);
            }
            else {
                assert !(!this.pkg.files.containsAll(this.pkg.getClassStubs()));
                ArrayList files = this.pkg.files;
                if (!PackerImpl.DoPack.$assertionsDisabled) {
                    (files = new ArrayList(this.pkg.files)).retainAll(this.pkg.getClassStubs());
                }
                assert !(!files.equals(this.pkg.getClassStubs()));
            }
            this.pkg.trimStubs();
            if (PackerImpl.this.props.getBoolean("com.sun.java.util.jar.pack.strip.debug")) {
                this.pkg.stripAttributeKind("Debug");
            }
            if (PackerImpl.this.props.getBoolean("com.sun.java.util.jar.pack.strip.compile")) {
                this.pkg.stripAttributeKind("Compile");
            }
            if (PackerImpl.this.props.getBoolean("com.sun.java.util.jar.pack.strip.constants")) {
                this.pkg.stripAttributeKind("Constant");
            }
            if (PackerImpl.this.props.getBoolean("com.sun.java.util.jar.pack.strip.exceptions")) {
                this.pkg.stripAttributeKind("Exceptions");
            }
            if (PackerImpl.this.props.getBoolean("com.sun.java.util.jar.pack.strip.innerclasses")) {
                this.pkg.stripAttributeKind("InnerClasses");
            }
            PackageWriter o = new PackageWriter(this.pkg, outputStream);
            o.archiveNextCount = archiveNextCount;
            o.write();
            outputStream.flush();
            if (this.verbose > 0) {
                long lng = o.archiveSize0 + o.archiveSize1;
                this.totalOutputSize += lng;
                Utils.log.info("Transmitted " + size + " files of " + this.segmentSize + " input bytes in a segment of " + lng + " bytes");
            }
        }
        
        List<PackerImpl.DoPack.InFile> scanJar(JarFile jarFile) throws IOException {
            ArrayList<Object> list = new ArrayList();
            try {
                for (JarEntry jarEntry : Collections.list(jarFile.entries())) {
                    PackerImpl.DoPack.InFile o = new PackerImpl.DoPack.InFile(jarFile, jarEntry);
                    assert jarEntry.isDirectory() == o.name.endsWith("/");
                    list.add(o);
                }
            }
            catch (IllegalStateException cause) {
                throw new IOException(cause.getLocalizedMessage(), cause);
            }
            return list;
        }
        
        final class InFile
        {
            final String name;
            final JarFile jf;
            final JarEntry je;
            final File f;
            int modtime;
            int options;
            
            InFile(String pathname) {
                this.modtime = 0;
                this.name = Utils.getJarEntryName(pathname);
                this.f = new File(pathname);
                this.jf = null;
                this.je = null;
                int modtime = this.getModtime(this.f.lastModified());
                if (PackerImpl.DoPack.this.keepModtime && modtime != 0) {
                    this.modtime = modtime;
                }
                else if (PackerImpl.DoPack.this.latestModtime && modtime > PackerImpl.DoPack.this.pkg.default_modtime) {
                    PackerImpl.DoPack.this.pkg.default_modtime = modtime;
                }
            }
            
            InFile(JarFile jf, JarEntry je) {
                this.modtime = 0;
                this.name = Utils.getJarEntryName(je.getName());
                this.f = null;
                this.jf = jf;
                this.je = je;
                int modtime = this.getModtime(je.getTime());
                if (PackerImpl.DoPack.this.keepModtime && modtime != 0) {
                    this.modtime = modtime;
                }
                else if (PackerImpl.DoPack.this.latestModtime && modtime > PackerImpl.DoPack.this.pkg.default_modtime) {
                    PackerImpl.DoPack.this.pkg.default_modtime = modtime;
                }
                if (PackerImpl.DoPack.this.keepDeflateHint && je.getMethod() == 8) {
                    this.options |= 0x1;
                }
            }
            
            InFile(PackerImpl.DoPack doPack, JarEntry jarEntry) {
                this(doPack, null, jarEntry);
            }
            
            long getInputLength() {
                long n = (this.je != null) ? this.je.getSize() : this.f.length();
                assert n >= 0L : this + ".len=" + n;
                return Math.max(0L, n) + (long)this.name.length() + 5L;
            }
            
            int getModtime(long n) {
                long n2 = (n + 500L) / 1000L;
                if ((long)(int)n2 == n2) {
                    return (int)n2;
                }
                Utils.log.warning("overflow in modtime for " + this.f);
                return 0;
            }
            
            void copyTo(Package.File file) {
                if (this.modtime != 0) {
                    file.modtime = this.modtime;
                }
                file.options |= this.options;
            }
            
            InputStream getInputStream() throws IOException {
                if (this.jf != null) {
                    return this.jf.getInputStream(this.je);
                }
                return new FileInputStream(this.f);
            }
            
            @Override
            public String toString() {
                return this.name;
            }
        }
    }
}
